from __future__ import annotations

import json
import logging
import os

from angrmanagement.utils.env import app_root, is_pyinstaller

_l = logging.getLogger(name=__name__)


class LibraryDocs:
    """
    Implements the manager of library docs.
    """

    def __init__(self) -> None:
        self.func_docs = []

    def load_func_docs(self, path) -> None:
        if not os.path.isabs(path):
            path = os.path.join(app_root(), path) if is_pyinstaller() else os.path.join(app_root(), "..", path)
        path = os.path.normpath(path)
        _l.info("Loading library docs from %s.", path)
        docs = []
        if os.path.isdir(path):
            for filename in os.listdir(path):
                if filename.endswith(".json"):
                    jpath = os.path.join(path, filename)
                    with open(jpath) as jfile:
                        data = json.load(jfile)
                        docs.append(data)

        self.func_docs = docs

    def get_docstring_for_func_name(self, func_name: str):
        for library in self.func_docs:
            for func_dict in library:
                if "name" not in func_dict:
                    continue
                if "description" not in func_dict:
                    continue
                names = func_dict["name"]
                name_list = names.split(",")
                for name in name_list:
                    name = name.strip()
                    if func_name == name:
                        doc_string = func_dict["description"]
                        url = "http://"
                        ftype = "<>"
                        if "url" in func_dict:
                            url = func_dict["url"]
                        if "type" in func_dict:
                            ftype = func_dict["type"]
                        return doc_string, url, ftype
        return None



from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QPointF, Qt

from angrmanagement.utils.graph_layouter import GraphLayouter

from .qgraph import QZoomableDraggableGraphicsView
from .qgraph_arrow import QProximityGraphArrow

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.views.proximity_view import ProximityView
    from angrmanagement.ui.widgets.qproximitygraph_block import QProximityGraphBlock
    from angrmanagement.utils.edge import Edge


_l = logging.getLogger(name=__name__)


class QProximityGraph(QZoomableDraggableGraphicsView):
    LEFT_PADDING = 2000
    TOP_PADDING = 2000

    def __init__(self, instance: Instance, proximity_view: ProximityView, parent=None) -> None:
        super().__init__(parent=parent)

        self._instance = instance
        self._proximity_view = proximity_view

        self._graph = None
        self.blocks = set()
        self._edges: list[Edge] = []
        self._arrows_by_src: dict[Any, list[QProximityGraphArrow]] = defaultdict(list)
        self._arrows_by_dst: dict[Any, list[QProximityGraphArrow]] = defaultdict(list)
        self._arrows: list[QProximityGraphArrow] = []

    @property
    def graph(self):
        return self._graph

    @graph.setter
    def graph(self, v) -> None:
        if v is not self._graph:
            self._graph = v
            self.reload()

    def reload(self) -> None:
        self.request_relayout()

    def refresh(self) -> None:
        for block in self.blocks:
            block.refresh()
        scene = self.scene()
        if scene is not None:
            scene.update(self.sceneRect())

    def request_relayout(self) -> None:
        self._reset_scene()
        if self.graph is None:
            return

        # remove all arrows
        self._arrows.clear()
        self._arrows_by_src.clear()
        self._arrows_by_dst.clear()

        # remove all nodes
        self.blocks.clear()

        node_sizes = {}
        for node in self.graph.nodes():
            self.blocks.add(node)
            node_sizes[node] = (node.width, node.height)
        # gl = TreeGraphLayouter(self.graph, node_sizes, vertical=True, horizontal_spacing=5, vertical_spacing=30,
        #                        layer_sorter=self._sort_nodes,
        #                        )
        gl = GraphLayouter(
            self.graph, node_sizes, node_sorter=self._sort_nodes, x_margin=2, y_margin=2, row_margin=2, col_margin=2
        )

        self._edges = gl.edges

        min_x, max_x, min_y, max_y = 0, 0, 0, 0

        scene = self.scene()
        for node, (x, y) in gl.node_coordinates.items():
            scene.addItem(node)
            node.setPos(x, y)
            min_x = min(min_x, node.x())
            max_x = max(max_x, node.x() + node.width)
            min_y = min(min_y, node.y())
            max_y = max(max_y, node.y() + node.height)

        # draw edges
        for edge in self._edges:
            arrow = QProximityGraphArrow(self._proximity_view, edge, arrow_location="end", arrow_direction="down")
            self._arrows.append(arrow)
            self._arrows_by_src[edge.src].append(arrow)
            self._arrows_by_dst[edge.dst].append(arrow)
            scene.addItem(arrow)
            arrow.setPos(QPointF(*edge.coordinates[0]))

        min_x -= self.LEFT_PADDING
        max_x += self.LEFT_PADDING
        min_y -= self.TOP_PADDING
        max_y += self.TOP_PADDING

        self._reset_view()

    def _initial_position(self):
        ibr = self.scene().itemsBoundingRect()
        return ibr.center()

    def _sort_nodes(self, nodes: list[QProximityGraphBlock]) -> list[QProximityGraphBlock]:
        # Sort nodes based on the address of their first ref_at address in each function
        sorted_nodes = sorted(nodes, key=lambda x: next(iter(x._node.ref_at)) if x._node.ref_at else 0)
        return sorted_nodes

    #
    # Event handlers
    #

    def keyPressEvent(self, event) -> None:
        """

        :param QKeyEvent event:
        :return:
        """

        key = event.key()

        if key == Qt.Key_Tab:
            self._symexec_view.switch_to_disassembly_view()
            event.accept()

        super().keyPressEvent(event)

    def on_block_hovered(self, block) -> None:
        if block is None:
            return
        scene = self.scene()
        # move all relevant arrows to the top
        if block in self._arrows_by_src:
            arrows = set(self._arrows_by_src[block])
            for arrow in arrows:
                for item in scene.collidingItems(arrow):
                    if item in arrows:
                        continue
                    item.stackBefore(arrow)

        if block in self._arrows_by_dst:
            arrows = set(self._arrows_by_dst[block])
            for arrow in arrows:
                for item in scene.collidingItems(arrow):
                    if item in arrows:
                        continue
                    item.stackBefore(arrow)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from PySide6.QtWidgets import QComboBox, QFileDialog, QFrame, QHBoxLayout, QLabel, QPushButton

from angrmanagement.ui.menus.disasm_options_menu import DisasmOptionsMenu
from angrmanagement.ui.toolbars import NavToolbar

from .qdisasm_base_control import DisassemblyLevel
from .qdisasm_graph import QDisassemblyGraph
from .qlinear_viewer import QLinearDisassembly

if TYPE_CHECKING:
    from angr.knowledge_plugins import Function


class QDisasmStatusBar(QFrame):
    """
    Status and control bar for disassembly views
    """

    def __init__(self, disasm_view, parent=None) -> None:
        super().__init__(parent)

        self.disasm_view = disasm_view

        # widgets
        self._nav_toolbar: NavToolbar | None = None
        self._function_label: QLabel = None
        self._options_menu: DisasmOptionsMenu = None
        self._view_combo: QComboBox = None

        # information
        self._function: Function | None = None

        self._init_menu()
        self._init_widgets()

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, f) -> None:
        self._function = f
        self._update_function_label()

    #
    # Initialization
    #

    def _init_widgets(self) -> None:
        self._nav_toolbar = NavToolbar(
            self.disasm_view.jump_history,
            self.disasm_view.jump_back,
            self.disasm_view.jump_forward,
            self.disasm_view.jump_to_history_position,
            True,
            self,
        )

        # current function
        self._function_label = QLabel()

        self._view_combo = QComboBox(self)
        self._view_combo.addItem("Linear Disassembly", QLinearDisassembly)
        self._view_combo.addItem("Graph Disassembly", QDisassemblyGraph)
        self._view_combo.activated.connect(self._view_combo_changed)
        self.disasm_view.view_visibility_changed.connect(self._update_view_combo)
        self._update_view_combo()

        self._disasm_level_combo = QComboBox(self)
        self._disasm_level_combo.addItem("Machine Code", DisassemblyLevel.MachineCode)
        self._disasm_level_combo.addItem("Lifter IR", DisassemblyLevel.LifterIR)
        self._disasm_level_combo.addItem("AIL", DisassemblyLevel.AIL)
        self._disasm_level_combo.activated.connect(self._disasm_level_combo_changed)
        self.disasm_view.disassembly_level_changed.connect(self._update_disasm_level_combo)
        self._update_disasm_level_combo()

        # options button
        option_btn = QPushButton()
        option_btn.setText("Options")
        option_btn.setMenu(self._options_menu.qmenu())

        # Save image button
        saveimage_btn = QPushButton()
        saveimage_btn.setText("Save image...")
        saveimage_btn.clicked.connect(self._on_saveimage_btn_clicked)

        layout = QHBoxLayout()
        layout.setContentsMargins(3, 3, 3, 3)
        layout.setSpacing(3)

        layout.addWidget(self._nav_toolbar.qtoolbar())
        layout.addWidget(self._function_label)

        layout.addStretch(0)

        layout.addWidget(saveimage_btn)
        layout.addWidget(self._view_combo)
        layout.addWidget(self._disasm_level_combo)
        layout.addWidget(option_btn)

        self.setLayout(layout)

    def _init_menu(self) -> None:
        self._options_menu = DisasmOptionsMenu(self.disasm_view)

    #
    # Private methods
    #

    def _view_combo_changed(self, index: int) -> None:
        {
            QLinearDisassembly: self.disasm_view.display_linear_viewer,
            QDisassemblyGraph: self.disasm_view.display_disasm_graph,
        }[self._view_combo.itemData(index)]()

    def _update_view_combo(self) -> None:
        graph_type = type(self.disasm_view.current_graph)
        index = self._view_combo.findData(graph_type)
        self._view_combo.setCurrentIndex(index)

    def _disasm_level_combo_changed(self, index: int) -> None:
        new_level = self._disasm_level_combo.itemData(index)
        self.disasm_view.set_disassembly_level(new_level)

    def _update_disasm_level_combo(self) -> None:
        new_level = self.disasm_view.disassembly_level
        index = self._disasm_level_combo.findData(new_level)
        self._disasm_level_combo.setCurrentIndex(index)

    def _update_function_label(self) -> None:
        s = f"{self._function.name} ({self._function.addr:x})" if self._function is not None else ""
        self._function_label.setText(s)

    def _on_saveimage_btn_clicked(self) -> None:
        filename, folder = QFileDialog.getSaveFileName(self, "Save image...", "", "PNG Files (*.png);;Bitmaps (*.bmp)")
        if not filename or not folder:
            return

        self.disasm_view.save_image_to(os.path.join(folder, filename))



# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from base64 import b64decode
import os
import re
from warnings import warn

from qtpy import QtCore, QtGui, QtWidgets

from traitlets import Bool
from pygments.util import ClassNotFound

from .svg import save_svg, svg_to_clipboard, svg_to_image
from .jupyter_widget import JupyterWidget
from .styles import get_colors

try:
    from IPython.lib.latextools import latex_to_png
except ImportError:
    latex_to_png = None


def _ensure_dir_exists(path, mode=0o755):
    """ensure that a directory exists

    If it doesn't exists, try to create it and protect against a race condition
    if another process is doing the same.

    The default permissions are 755, which differ from os.makedirs default of 777.
    """
    if not os.path.exists(path):
        try:
            os.makedirs(path, mode=mode)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    elif not os.path.isdir(path):
        raise IOError("%r exists but is not a directory" % path)

class LatexError(Exception):
    """Exception for Latex errors"""


class RichIPythonWidget(JupyterWidget):
    """Dummy class for config inheritance. Destroyed below."""


class RichJupyterWidget(RichIPythonWidget):
    """ An JupyterWidget that supports rich text, including lists, images, and
        tables. Note that raw performance will be reduced compared to the plain
        text version.
    """

    # RichJupyterWidget protected class variables.
    _payload_source_plot = 'ipykernel.pylab.backend_payload.add_plot_payload'
    _jpg_supported = Bool(False)

    # Used to determine whether a given html export attempt has already
    # displayed a warning about being unable to convert a png to svg.
    _svg_warning_displayed = False

    #---------------------------------------------------------------------------
    # 'object' interface
    #---------------------------------------------------------------------------

    def __init__(self, *args, **kw):
        """ Create a RichJupyterWidget.
        """
        kw['kind'] = 'rich'
        super().__init__(*args, **kw)

        # Configure the ConsoleWidget HTML exporter for our formats.
        self._html_exporter.image_tag = self._get_image_tag

        # Dictionary for resolving document resource names to SVG data.
        self._name_to_svg_map = {}

        # Do we support jpg ?
        # it seems that sometime jpg support is a plugin of QT, so try to assume
        # it is not always supported.
        self._jpg_supported = 'jpeg' in QtGui.QImageReader.supportedImageFormats()


    #---------------------------------------------------------------------------
    # 'ConsoleWidget' public interface overides
    #---------------------------------------------------------------------------

    def export_html(self):
        """ Shows a dialog to export HTML/XML in various formats.

        Overridden in order to reset the _svg_warning_displayed flag prior
        to the export running.
        """
        self._svg_warning_displayed = False
        super().export_html()


    #---------------------------------------------------------------------------
    # 'ConsoleWidget' protected interface
    #---------------------------------------------------------------------------

    def _context_menu_make(self, pos):
        """ Reimplemented to return a custom context menu for images.
        """
        format = self._control.cursorForPosition(pos).charFormat()
        name = format.stringProperty(QtGui.QTextFormat.ImageName)
        if name:
            menu = QtWidgets.QMenu(self)

            menu.addAction('Copy Image', lambda: self._copy_image(name))
            menu.addAction('Save Image As...', lambda: self._save_image(name))
            menu.addSeparator()

            svg = self._name_to_svg_map.get(name, None)
            if svg is not None:
                menu.addSeparator()
                menu.addAction('Copy SVG', lambda: svg_to_clipboard(svg))
                menu.addAction('Save SVG As...',
                               lambda: save_svg(svg, self._control))
        else:
            menu = super()._context_menu_make(pos)
        return menu

    #---------------------------------------------------------------------------
    # 'BaseFrontendMixin' abstract interface
    #---------------------------------------------------------------------------
    def _pre_image_append(self, msg, prompt_number):
        """Append the Out[] prompt  and make the output nicer

        Shared code for some the following if statement
        """
        self._append_plain_text(self.output_sep, True)
        self._append_html(self._make_out_prompt(prompt_number), True)
        self._append_plain_text('\n', True)

    def _handle_execute_result(self, msg):
        """Overridden to handle rich data types, like SVG."""
        self.log.debug("execute_result: %s", msg.get('content', ''))
        if self.include_output(msg):
            self.flush_clearoutput()
            content = msg['content']
            prompt_number = content.get('execution_count', 0)
            data = content['data']
            metadata = msg['content']['metadata']
            if 'image/svg+xml' in data:
                self._pre_image_append(msg, prompt_number)
                self._append_svg(data['image/svg+xml'], True)
                self._append_html(self.output_sep2, True)
            elif 'image/png' in data:
                self._pre_image_append(msg, prompt_number)
                png = b64decode(data['image/png'].encode('ascii'))
                self._append_png(png, True, metadata=metadata.get('image/png',
                                                                  None))
                self._append_html(self.output_sep2, True)
            elif 'image/jpeg' in data and self._jpg_supported:
                self._pre_image_append(msg, prompt_number)
                jpg = b64decode(data['image/jpeg'].encode('ascii'))
                self._append_jpg(jpg, True, metadata=metadata.get('image/jpeg',
                                                                  None))
                self._append_html(self.output_sep2, True)
            elif 'text/latex' in data:
                self._pre_image_append(msg, prompt_number)
                try:
                    self._append_latex(data['text/latex'], True)
                except LatexError:
                    return super()._handle_display_data(msg)
                self._append_html(self.output_sep2, True)
            else:
                # Default back to the plain text representation.
                return super()._handle_execute_result(msg)

    def _handle_display_data(self, msg):
        """Overridden to handle rich data types, like SVG."""
        self.log.debug("display_data: %s", msg.get('content', ''))
        if self.include_output(msg):
            self.flush_clearoutput()
            data = msg['content']['data']
            metadata = msg['content']['metadata']
            # Try to use the svg or html representations.
            # FIXME: Is this the right ordering of things to try?
            self.log.debug("display: %s", msg.get('content', ''))
            if 'image/svg+xml' in data:
                svg = data['image/svg+xml']
                self._append_svg(svg, True)
            elif 'image/png' in data:
                # PNG data is base64 encoded as it passes over the network
                # in a JSON structure so we decode it.
                png = b64decode(data['image/png'].encode('ascii'))
                self._append_png(png, True, metadata=metadata.get('image/png', None))
            elif 'image/jpeg' in data and self._jpg_supported:
                jpg = b64decode(data['image/jpeg'].encode('ascii'))
                self._append_jpg(jpg, True, metadata=metadata.get('image/jpeg', None))
            elif 'text/latex' in data and latex_to_png:
                try:
                    self._append_latex(data['text/latex'], True)
                except LatexError:
                    return super()._handle_display_data(msg)
            else:
                # Default back to the plain text representation.
                return super()._handle_display_data(msg)

    #---------------------------------------------------------------------------
    # 'RichJupyterWidget' protected interface
    #---------------------------------------------------------------------------
    def _is_latex_math(self, latex):
        """
        Determine if a Latex string is in math mode

        This is the only mode supported by qtconsole
        """
        basic_envs = ['math', 'displaymath']
        starable_envs = ['equation', 'eqnarray' 'multline', 'gather', 'align',
                         'flalign', 'alignat']
        star_envs = [env + '*' for env in starable_envs]
        envs = basic_envs + starable_envs + star_envs

        env_syntax = [r'\begin{{{0}}} \end{{{0}}}'.format(env).split() for env in envs]

        math_syntax = [
            (r'\[', r'\]'), (r'\(', r'\)'),
            ('$$', '$$'), ('$', '$'),
        ]

        for start, end in math_syntax + env_syntax:
            inner = latex[len(start):-len(end)]
            if start in inner or end in inner:
                return False
            if latex.startswith(start) and latex.endswith(end):
                return True
        return False

    def _get_color(self, color):
        """Get color from the current syntax style if loadable."""
        try:
            return get_colors(self.syntax_style)[color]
        except ClassNotFound:
            # The syntax_style has been sideloaded (e.g. by spyder).
            # In this case the overloading class should override this method.
            return get_colors('default')[color]

    def _append_latex(self, latex, before_prompt=False, metadata=None):
        """ Append latex data to the widget."""
        png = None

        if self._is_latex_math(latex):
            png = latex_to_png(latex, wrap=False, backend='dvipng',
                               color=self._get_color('fgcolor'))

        # Matplotlib only supports strings enclosed in dollar signs
        if png is None and latex.startswith('$') and latex.endswith('$'):
            # To avoid long and ugly errors, like the one reported in
            # spyder-ide/spyder#7619
            try:
                png = latex_to_png(latex, wrap=False, backend='matplotlib',
                                   color=self._get_color('fgcolor'))
            except Exception:
                pass

        if png:
            self._append_png(png, before_prompt, metadata)
        else:
            raise LatexError

    def _append_jpg(self, jpg, before_prompt=False, metadata=None):
        """ Append raw JPG data to the widget."""
        self._append_custom(self._insert_jpg, jpg, before_prompt, metadata=metadata)

    def _append_png(self, png, before_prompt=False, metadata=None):
        """ Append raw PNG data to the widget.
        """
        self._append_custom(self._insert_png, png, before_prompt, metadata=metadata)

    def _append_svg(self, svg, before_prompt=False):
        """ Append raw SVG data to the widget.
        """
        self._append_custom(self._insert_svg, svg, before_prompt)

    def _add_image(self, image):
        """ Adds the specified QImage to the document and returns a
            QTextImageFormat that references it.
        """
        document = self._control.document()
        name = str(image.cacheKey())
        document.addResource(QtGui.QTextDocument.ImageResource,
                             QtCore.QUrl(name), image)
        format = QtGui.QTextImageFormat()
        format.setName(name)
        return format

    def _copy_image(self, name):
        """ Copies the ImageResource with 'name' to the clipboard.
        """
        image = self._get_image(name)
        QtWidgets.QApplication.clipboard().setImage(image)

    def _get_image(self, name):
        """ Returns the QImage stored as the ImageResource with 'name'.
        """
        document = self._control.document()
        image = document.resource(QtGui.QTextDocument.ImageResource,
                                  QtCore.QUrl(name))
        return image

    def _get_image_tag(self, match, path = None, format = "png"):
        """ Return (X)HTML mark-up for the image-tag given by match.

        Parameters
        ----------
        match : re.SRE_Match
            A match to an HTML image tag as exported by Qt, with
            match.group("Name") containing the matched image ID.

        path : string|None, optional [default None]
            If not None, specifies a path to which supporting files may be
            written (e.g., for linked images).  If None, all images are to be
            included inline.

        format : "png"|"svg"|"jpg", optional [default "png"]
            Format for returned or referenced images.
        """
        if format in ("png","jpg"):
            try:
                image = self._get_image(match.group("name"))
            except KeyError:
                return "<b>Couldn't find image %s</b>" % match.group("name")

            if path is not None:
                _ensure_dir_exists(path)
                relpath = os.path.basename(path)
                if image.save("%s/qt_img%s.%s" % (path, match.group("name"), format),
                              "PNG"):
                    return '<img src="%s/qt_img%s.%s">' % (relpath,
                                                            match.group("name"),format)
                else:
                    return "<b>Couldn't save image!</b>"
            else:
                ba = QtCore.QByteArray()
                buffer_ = QtCore.QBuffer(ba)
                buffer_.open(QtCore.QIODevice.WriteOnly)
                image.save(buffer_, format.upper())
                buffer_.close()
                return '<img src="data:image/%s;base64,\n%s\n" />' % (
                    format,re.sub(r'(.{60})',r'\1\n', str(ba.toBase64().data().decode())))

        elif format == "svg":
            try:
                svg = str(self._name_to_svg_map[match.group("name")])
            except KeyError:
                if not self._svg_warning_displayed:
                    QtWidgets.QMessageBox.warning(self, 'Error converting PNG to SVG.',
                        'Cannot convert PNG images to SVG, export with PNG figures instead. '
                        'If you want to export matplotlib figures as SVG, add '
                        'to your ipython config:\n\n'
                        '\tc.InlineBackend.figure_format = \'svg\'\n\n'
                        'And regenerate the figures.',
                                              QtWidgets.QMessageBox.Ok)
                    self._svg_warning_displayed = True
                return ("<b>Cannot convert  PNG images to SVG.</b>  "
                        "You must export this session with PNG images. "
                        "If you want to export matplotlib figures as SVG, add to your config "
                        "<span>c.InlineBackend.figure_format = 'svg'</span> "
                        "and regenerate the figures.")

            # Not currently checking path, because it's tricky to find a
            # cross-browser way to embed external SVG images (e.g., via
            # object or embed tags).

            # Chop stand-alone header from matplotlib SVG
            offset = svg.find("<svg")
            assert(offset > -1)

            return svg[offset:]

        else:
            return '<b>Unrecognized image format</b>'

    def _insert_jpg(self, cursor, jpg, metadata=None):
        """ Insert raw PNG data into the widget."""
        self._insert_img(cursor, jpg, 'jpg', metadata=metadata)

    def _insert_png(self, cursor, png, metadata=None):
        """ Insert raw PNG data into the widget.
        """
        self._insert_img(cursor, png, 'png', metadata=metadata)

    def _insert_img(self, cursor, img, fmt, metadata=None):
        """ insert a raw image, jpg or png """
        if metadata:
            width = metadata.get('width', None)
            height = metadata.get('height', None)
        else:
            width = height = None
        try:
            image = QtGui.QImage()
            image.loadFromData(img, fmt.upper())
            if width and height:
                image = image.scaled(int(width), int(height),
                                     QtCore.Qt.IgnoreAspectRatio,
                                     QtCore.Qt.SmoothTransformation)
            elif width and not height:
                image = image.scaledToWidth(int(width), QtCore.Qt.SmoothTransformation)
            elif height and not width:
                image = image.scaledToHeight(int(height), QtCore.Qt.SmoothTransformation)
        except ValueError:
            self._insert_plain_text(cursor, 'Received invalid %s data.'%fmt)
        else:
            format = self._add_image(image)
            cursor.insertBlock()
            cursor.insertImage(format)
            cursor.insertBlock()

    def _insert_svg(self, cursor, svg):
        """ Insert raw SVG data into the widet.
        """
        try:
            image = svg_to_image(svg)
        except ValueError:
            self._insert_plain_text(cursor, 'Received invalid SVG data.')
        else:
            format = self._add_image(image)
            self._name_to_svg_map[format.name()] = svg
            cursor.insertBlock()
            cursor.insertImage(format)
            cursor.insertBlock()

    def _save_image(self, name, format='PNG'):
        """ Shows a save dialog for the ImageResource with 'name'.
        """
        dialog = QtWidgets.QFileDialog(self._control, 'Save Image')
        dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptSave)
        dialog.setDefaultSuffix(format.lower())
        dialog.setNameFilter('%s file (*.%s)' % (format, format.lower()))
        if dialog.exec_():
            filename = dialog.selectedFiles()[0]
            image = self._get_image(name)
            image.save(filename, format)


# Clobber RichIPythonWidget above:

class RichIPythonWidget(RichJupyterWidget):
    """Deprecated class. Use RichJupyterWidget."""
    def __init__(self, *a, **kw):
        warn("RichIPythonWidget is deprecated, use RichJupyterWidget",
             DeprecationWarning)
        super().__init__(*a, **kw)



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from PySide6.QtCore import QPointF, Qt
from PySide6.QtGui import QBrush, QColor, QKeyEvent, QPainterPath, QPainterPathStroker, QPen
from PySide6.QtWidgets import QApplication, QGraphicsItem

from angrmanagement.config import Conf
from angrmanagement.utils.edge import EdgeSort

if TYPE_CHECKING:
    from angrmanagement.ui.views.data_dep_view import DataDepView
    from angrmanagement.ui.views.dep_view import DependencyView
    from angrmanagement.ui.views.proximity_view import ProximityView

EDGE_COLORS = {
    EdgeSort.BACK_EDGE: "disasm_view_back_edge_color",
    EdgeSort.TRUE_BRANCH: "disasm_view_true_edge_color",
    EdgeSort.FALSE_BRANCH: "disasm_view_false_edge_color",
    EdgeSort.DIRECT_JUMP: "disasm_view_direct_jump_edge_color",
    EdgeSort.EXCEPTION_EDGE: "disasm_view_exception_edge_color",
}

EDGE_STYLES = {EdgeSort.DIRECT_JUMP: Qt.SolidLine, EdgeSort.EXCEPTION_EDGE: Qt.DashLine}


class QGraphArrow(QGraphicsItem):
    def __init__(self, edge, arrow_location: str = "end", arrow_direction: str = "down", parent=None) -> None:
        super().__init__(parent)

        self.edge = edge
        self.rect = None
        self._start = QPointF(*self.edge.coordinates[0])
        self.coords = [self.create_point(c) for c in self.edge.coordinates]
        self.end = self.coords[-1]
        self.start = self.coords[0]

        self.color = getattr(Conf, EDGE_COLORS.get(self.edge.sort, EDGE_COLORS[EdgeSort.DIRECT_JUMP]))
        self.arrow = self._make_arrow(arrow_location, arrow_direction)
        self.style = EDGE_STYLES.get(self.edge.sort, EDGE_STYLES[EdgeSort.DIRECT_JUMP])
        self.path = self._make_path()

        self._hovered = False

        self.setAcceptHoverEvents(True)

    @staticmethod
    def _get_distance(pt0: QPointF, pt1: QPointF) -> float:
        d = (pt1.x() - pt0.x()) ** 2 + (pt1.y() - pt0.y()) ** 2
        return math.sqrt(d)

    def _make_path(self):
        path = QPainterPath(self.coords[0])
        for c in self.coords[1:]:
            path.lineTo(c)
        return path

    def _make_arrow(self, location, direction):
        coord = self.start if location == "start" else self.end

        if direction == "down":
            return [
                QPointF(coord.x() - 3, coord.y()),
                QPointF(coord.x() + 3, coord.y()),
                QPointF(coord.x(), coord.y() + 6),
            ]
        elif direction == "right":
            return [
                QPointF(coord.x(), coord.y() - 3),
                QPointF(coord.x(), coord.y() + 3),
                QPointF(coord.x() + 6, coord.y()),
            ]
        elif direction == "left":
            return [
                QPointF(coord.x(), coord.y() - 3),
                QPointF(coord.x(), coord.y() + 3),
                QPointF(coord.x() - 6, coord.y()),
            ]
        else:
            raise NotImplementedError(f"Direction {direction} is not supported yet.")

    def _should_highlight(self) -> bool:
        return False

    def create_point(self, stuff):
        return QPointF(*stuff) - self._start

    def paint(self, painter, option, widget) -> None:
        lod = option.levelOfDetailFromTransform(painter.worldTransform())
        should_highlight = self._should_highlight()

        pen = QPen(QColor(0, 254, 254), 2, self.style) if should_highlight else QPen(self.color, 2, self.style)

        painter.setPen(pen)

        painter.drawPath(self.path)

        # arrow
        if lod < 0.3:
            return

        # arrow
        brush = QBrush(QColor(0, 254, 254)) if should_highlight else QBrush(self.color)
        painter.setBrush(brush)
        painter.drawPolygon(self.arrow)

    def boundingRect(self):
        if self.rect is None:
            self.rect = self.path.boundingRect()
        return self.rect

    def shape(self):
        stroker = QPainterPathStroker()
        stroker.setWidth(4)
        stroker.setCapStyle(Qt.RoundCap)
        return stroker.createStroke(self.path)

    #
    # Event handlers
    #

    def hoverEnterEvent(self, event) -> None:
        pass

    def hoverLeaveEvent(self, event) -> None:
        pass

    def mouseDoubleClickEvent(self, event) -> None:
        pass


class QDisasmGraphArrow(QGraphArrow):
    def __init__(self, edge, disasm_view, infodock, parent=None) -> None:
        super().__init__(edge, arrow_direction="down", parent=parent)
        self.disasm_view = disasm_view
        self.infodock = infodock

    def _should_highlight(self) -> bool:
        return (
            self.infodock.is_edge_hovered(self.edge.src.addr, self.edge.dst.addr)
            or self.infodock.is_block_hovered(self.edge.src.addr)
            or self.infodock.is_block_hovered(self.edge.dst.addr)
        )

    #
    # Event handlers
    #

    def hoverEnterEvent(self, event) -> None:
        self.infodock.hover_edge(self.edge.src.addr, self.edge.dst.addr)

    def hoverLeaveEvent(self, event) -> None:
        self.infodock.unhover_edge(self.edge.src.addr, self.edge.dst.addr)

    def mouseDoubleClickEvent(self, event) -> None:
        if QApplication.keyboardModifiers() == Qt.ShiftModifier:
            # go to the source
            self.disasm_view.jump_to(self.edge.src.addr, src_ins_addr=self.edge.dst.addr, use_animation=True)
            event.accept()
        else:
            # go to the destination
            self.disasm_view.jump_to(self.edge.dst.addr, src_ins_addr=self.edge.src.addr, use_animation=True)
            event.accept()
        super().mouseDoubleClickEvent(event)


class QGraphArrowBezier(QGraphArrow):
    def __init__(
        self, edge, arrow_location: str = "end", arrow_direction: str = "down", radius: int = 18, parent=None
    ) -> None:
        self._radius = radius
        super().__init__(edge, arrow_location=arrow_location, arrow_direction=arrow_direction, parent=parent)

    def _get_line_start(self, i: int) -> QPointF:
        pt0 = self.coords[i]
        pt1 = self.coords[i + 1] if i + 1 < len(self.coords) else self.coords[0]
        rat = self._radius / self._get_distance(pt0, pt1)
        if rat > 0.5:
            rat = 0.5
        return QPointF((1.0 - rat) * pt0.x() + rat * pt1.x(), (1.0 - rat) * pt0.y() + rat * pt1.y())

    def _get_line_end(self, i: int) -> QPointF:
        pt0 = self.coords[i]
        pt1 = self.coords[i + 1] if i + 1 < len(self.coords) else self.coords[0]
        rat = self._radius / self._get_distance(pt0, pt1)
        if rat > 0.5:
            rat = 0.5
        return QPointF(rat * pt0.x() + (1.0 - rat) * pt1.x(), rat * pt0.y() + (1.0 - rat) * pt1.y())

    def _make_path(self):
        if len(self.coords) < 3:
            return super()._make_path()
            # programming error - don't use this class for a simple segment!
            # raise ValueError("At least 3 coordinates are required.")

        path = QPainterPath(self.coords[0])

        for i in range(len(self.coords) - 1):
            pt0 = self._get_line_start(i)
            if i == 0:
                path.lineTo(pt0)
            else:
                path.quadTo(self.coords[i], pt0)
            pt1 = self._get_line_end(i)
            path.lineTo(pt1)

        path.lineTo(self.coords[-1])

        return path


class QDepGraphArrow(QGraphArrowBezier):
    def __init__(self, dep_view: DependencyView, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._dep_view = dep_view

    def _should_highlight(self) -> bool:
        if self._dep_view.hovered_block is None:
            return False
        return self._dep_view.hovered_block is self.edge.src or self._dep_view.hovered_block is self.edge.dst


class QProximityGraphArrow(QGraphArrow):
    def __init__(self, proximity_view: ProximityView, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._proximity_view = proximity_view

    def _should_highlight(self) -> bool:
        if self._proximity_view.hovered_block is None:
            return False
        return (
            self._proximity_view.hovered_block is self.edge.src or self._proximity_view.hovered_block is self.edge.dst
        )


class QDataDepGraphArrow(QGraphArrow):
    """Used to represent an edge between two QDataDepGraphBlocks"""

    def __init__(self, data_dep_view: DataDepView, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setFlags(QGraphicsItem.ItemIsFocusable)
        self._data_dep_view = data_dep_view
        self.is_hovered = False

    def _should_highlight(self) -> bool:
        # Should be highlighted if in trace path
        return (
            self.is_hovered
            or self.edge.dst in self._data_dep_view.traced_ancestors
            or self.edge.src in self._data_dep_view.traced_descendants
        )

    def hoverEnterEvent(self, event) -> None:
        self.is_hovered = True
        self.update()  # Must trigger repaint to highlight
        self.setFocus(Qt.MouseFocusReason)
        self.grabKeyboard()
        self._data_dep_view.graph_widget.handle_preview_request(self, event.modifiers() & Qt.ControlModifier)

    def hoverLeaveEvent(self, event) -> None:
        self.is_hovered = False
        self.update()  # Must trigger repaint to unhighlight
        self.ungrabKeyboard()
        self.clearFocus()
        self._data_dep_view.graph_widget.hide_preview()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if self.is_hovered and event.key() == Qt.Key_Control:
            self._data_dep_view.graph_widget.handle_preview_request(self, True)
        else:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        if self.is_hovered and event.key() == Qt.Key_Control:
            self._data_dep_view.graph_widget.handle_preview_request(self, False)
        else:
            super().keyPressEvent(event)

    def mouseDoubleClickEvent(self, event) -> None:
        """
        A double click on an arrow should center the data-dep view on the arrow's destination or source node
        If the control modifier is held, then the jump will be made to the destination. Otherwise, the source.
        """
        jump_to_dst = event.modifiers() & Qt.ControlModifier
        self._data_dep_view.graph_widget.jump_to_neighbor(self, jump_to_dst, self.mapToParent(event.pos().toPoint()))


class QDataDepGraphAncestorLine(QDataDepGraphArrow):
    """
    Dashed line to differentiate between value dependencies and ancestor linking
    """

    dash_len = 5.0

    def _calculate_dash_pattern(self) -> list[float]:
        """
        Builds dash pattern list dynamically, ensuring the correct number of dashes and spacing per each distance that
        must be covered
        """
        dash_len = self.dash_len
        distance = self._get_distance(self.start, self.end)
        dash_count = distance // dash_len

        dash_count = 19 if dash_count < 20 else dash_count - 1
        dash_len = (distance - dash_len) / dash_count

        # Must have an even amount of dashes
        if dash_count % 2 != 0:
            dash_count -= 1
            dash_len = distance / dash_count

        pattern = [dash_len] * int(dash_count)
        pattern.append(distance - (dash_len * dash_count))  # Remaining width is last tick
        pattern.append(0.0)
        return pattern

    def paint(self, painter, option, widget) -> None:
        should_highlight = self._should_highlight()

        pen = QPen(QColor(0, 254, 254), 2, self.style) if should_highlight else QPen(self.color, 2, self.style)

        pen.setDashPattern(self._calculate_dash_pattern())

        painter.setPen(pen)
        painter.drawPath(self.path)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from PySide6.QtCore import QSize, QTimer
from PySide6.QtWidgets import QTextEdit, QVBoxLayout

from angrmanagement.ui.views import InstanceView

if TYPE_CHECKING:
    from binharness import IO

log = logging.getLogger(name=__name__)


class StreamWidget(QTextEdit):
    """StreamWidget displays a stream of bytes as text."""

    def __init__(self, stream: IO[bytes]):
        super().__init__()
        self.stream = stream
        self.stream.set_blocking(False)
        self.buffer = b""
        self.setReadOnly(True)
        self.setText("")

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.reload)
        self.timer.start(100)

    def reload(self):
        new_data = self.stream.read(1024)
        if new_data:
            self.buffer += new_data
            self.setText(self.buffer.decode("utf-8"))


class StreamView(InstanceView):
    """StreamView displays a stream of bytes as text."""

    _stream: IO[bytes]
    _buffer: str

    _text_edit: QTextEdit

    def __init__(self, workspace, instance, default_docking_position, stream: IO[bytes], caption: str):
        super().__init__("log", workspace, default_docking_position, instance)

        self.base_caption = caption

        log.debug("StreamView initializing")
        self._text_edit = StreamWidget(stream)

        hlayout = QVBoxLayout()
        hlayout.addWidget(self._text_edit)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)
        log.debug("StreamView initialized")

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint: disable=unused-argument
        return QSize(50, 0)



from __future__ import annotations

import asyncio
import json
import logging
import math
import threading
import time
from typing import TYPE_CHECKING

from PySide6.QtGui import QColor
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from angrmanagement.config import Conf
from angrmanagement.errors import UnexpectedStatusCodeError
from angrmanagement.logic.threads import gui_thread_schedule, gui_thread_schedule_async
from angrmanagement.plugins import BasePlugin
from angrmanagement.utils.io import download_url

from .parse_trace import trace_to_bb_addrs

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

log = logging.getLogger(__name__)

try:
    import slacrs.model
except ImportError:
    log.error("You don't have slacrs module installed locally, CoveragePlugin going to have a bad time.")


def generate_light_gradients(color, number: int, lightness: int = 20):
    """
    return a List of QColors, where the colors are ordered in terms of
    lightness (last is given as color) the rest (total of number) are
    each lightness (%) lighter.
    """
    to_return = [color]
    for _ in range(number):
        last_color = to_return[-1]
        to_return.append(last_color.lighter(100 + lightness))
    to_return.reverse()
    return to_return


# TODO: This should really be a property of the target
TRACE_BASE = 0x4000000000


class CoveragePlugin(BasePlugin):
    """
    Implements the fuzzing coverage view.
    """

    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)

        self.workspace = workspace

        # attempt to connect to the necessary components
        self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")

        if self.connector is None:
            self.workspace.log("Unable to retrieve plugin ChessConnector")

        self.slacrs_instance = self.connector.slacrs_instance()
        if self.slacrs_instance is None:
            self.workspace.log("Unable to retrieve Slacrs instance")

        self.dark_theme_color = QColor(0, 20, 147)
        self.light_theme_color = QColor(225, 174, 0)
        self.hit_color = self.dark_theme_color if Conf.theme_name == "dark" else self.light_theme_color
        self.num_gradients = 16
        self.gradients = generate_light_gradients(self.hit_color, self.num_gradients, lightness=int(100 / 16))

        self.running = False
        self.slacrs_thread = None

        self.seen_traces = None
        self.bbl_coverage = None
        self.bbl_coverage_hash = 0

        self.coverage_lock = threading.Lock()
        self.reset_coverage()

    MENU_BUTTONS = [
        "Start Showing Coverage",
        "Stop Showing Coverage",
    ]
    START_SHOWING_COVERAGE = 0
    STOP_SHOWING_COVERAGE = 1

    def handle_click_menu(self, idx: int) -> None:
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        mapping = {
            self.START_SHOWING_COVERAGE: self.start,
            self.STOP_SHOWING_COVERAGE: self.stop,
        }

        mapping.get(idx)()

    def start(self) -> None:
        self.running = True
        self.slacrs_thread = threading.Thread(target=self.listen_for_events, daemon=True)
        self.slacrs_thread.start()
        gui_thread_schedule(self._refresh_gui)

    def stop(self) -> None:
        self.running = False
        gui_thread_schedule(self._refresh_gui)
        if self.workspace._main_window is not None:
            gui_thread_schedule_async(self.workspace._main_window.progress_done)

    def _coverage_of_func(self, func):
        """
        return (set of covered_bbls, and num_of_function_bbls)
        """
        func_bbls = func.block_addrs_set
        with self.coverage_lock:
            covered_bbls = self.bbl_coverage & func_bbls

        return covered_bbls, len(func_bbls)

    def color_block(self, addr: int):
        if not self.running:
            return None
        with self.coverage_lock:
            if addr in self.bbl_coverage:
                return self.dark_theme_color if Conf.theme_name == "dark" else self.light_theme_color
        return None

    def color_func(self, func):
        if not self.running:
            return None

        covered_bbls, total_bbls = self._coverage_of_func(func)

        # Be paranoid
        if total_bbls == 0:
            return None

        # Never want to highlight something that wasn't covered
        if len(covered_bbls) == 0:
            return None

        fraction_covered = len(covered_bbls) / total_bbls

        gradient_number = math.ceil(fraction_covered * len(self.gradients))
        return self.gradients[gradient_number - 1]

    FUNC_COLUMNS = ("Fuzzing Coverage",)

    def extract_func_column(self, func, idx: int):
        assert idx == 0
        if not self.running:
            return 0, "0%"

        covered_bbls, total_bbls = self._coverage_of_func(func)
        if len(covered_bbls) == 0:
            return 0, "0%"

        fraction_covered = len(covered_bbls) / total_bbls

        return fraction_covered, f"{int(round(fraction_covered*100,0))}%"

    def _refresh_gui(self) -> None:
        self.workspace.refresh()

    def reset_coverage(self) -> None:
        with self.coverage_lock:
            self.seen_traces = set()
            self.bbl_coverage = set()
            self.bbl_coverage_hash = 0

    def update_coverage_from_list(self, trace_addrs) -> None:
        log.info("Processing %d from the trace", len(trace_addrs))
        with self.coverage_lock:
            for addr in trace_addrs:
                self.bbl_coverage.add(addr)

        new_hash = hash(frozenset(self.bbl_coverage))
        if new_hash != self.bbl_coverage_hash:
            self.bbl_coverage_hash = new_hash
            gui_thread_schedule(self._refresh_gui)

    def update_coverage(self) -> None:
        self.set_status("Retrieving fuzzing coverage information...", 0.0)
        session = self.slacrs_instance.session()
        if session:
            for idx, trace in enumerate(
                session.query(slacrs.model.Trace)
                .filter(slacrs.model.Trace.input.has(target_image_id=self.connector.target_image_id))
                .order_by(slacrs.model.Trace.created_at)
            ):
                if not self.running:
                    break
                self.set_status(f"Processing trace {idx}...", 50.0)
                self.update_one_coverage(trace)
        self.set_status("Fuzzing coverage updated", 100.0)

    def update_one_coverage(self, trace) -> None:
        with self.coverage_lock:
            if trace.id in self.seen_traces:
                log.info("Already seen trace %s, skipping", trace.id)
                return

        log.info("Processing trace %s %s %s", trace.id, trace.input_id, trace.created_at)

        if not Conf.checrs_rest_endpoint_url:
            log.error("Unable to fetch trace %d because there is no CHECRS REST endpoint.", trace.id)
            return

        url = f"{Conf.checrs_rest_endpoint_url}v1/targets/{self.connector.target_image_id}/seeds/{trace.input_id}/trace"
        try:
            trace_bytes = download_url(url, parent=self.workspace._main_window, to_file=False)
        except UnexpectedStatusCodeError:
            log.exception("Unable to download %s.", url)
            return
        try:
            parsed_trace = json.loads(trace_bytes)
        except json.JSONDecodeError:
            log.exception("Unable to parse %s as JSON.", url)
            return

        bbl_addrs = trace_to_bb_addrs(parsed_trace, self.workspace.main_instance.project, TRACE_BASE)
        self.update_coverage_from_list(bbl_addrs)

        with self.coverage_lock:
            self.seen_traces.add(trace.id)
        log.info("Done processing trace %s.", trace.id)

    def listen_for_events(self) -> None:
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        while not self.connector and self.running:
            self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
            time.sleep(1)

        while not self.slacrs_instance and self.running:
            self.slacrs_instance = self.connector.slacrs_instance()
            time.sleep(1)

        while not self.connector.target_image_id and self.running:
            time.sleep(1)

        if not self.running:
            return

        self.update_coverage()

        prev_target = self.connector.target_image_id
        while self.running:
            if self.connector.target_image_id != prev_target:
                self.reset_coverage()
                self.update_coverage()

            self.set_status("Retrieving fuzzing coverage information...", 0.0)
            new_event_count = self.slacrs_instance.fetch_events()
            trace_idx = 0
            for idx in range(new_event_count):
                e = self.slacrs_instance.event_queue.get_nowait()
                session = self.slacrs_instance.session()
                if e.kind == "trace":
                    obj = e.get_object(session)
                    if session.query(slacrs.model.Trace).filter_by(id=e.object_id) == 1:
                        if not self.running:
                            break
                        trace = session.query(slacrs.model.Trace).filter_by(obj.object_id).one()
                        self.set_status(f"Processing trace {trace_idx}...", idx * 100 / new_event_count)
                        trace_idx += 1
                        self.update_one_coverage(trace)
                session.close()
            self.set_status("Fuzzing coverage updated", 100.0)

    def set_status(self, status: str, percentage: float) -> None:
        if self.workspace._main_window is not None:
            gui_thread_schedule_async(self.workspace._main_window.progress, (status, percentage))



from __future__ import annotations

from IPython.lib import guisupport

from angrmanagement.ui.css import CSS
from angrmanagement.vendor.qtconsole.inprocess import QtInProcessKernelManager
from angrmanagement.vendor.qtconsole.rich_jupyter_widget import RichJupyterWidget


class QIPythonWidget(RichJupyterWidget):
    def __init__(self, *args, banner=None, namespace=None, **kwargs) -> None:
        if banner is not None:
            self.banner = banner
        RichJupyterWidget.__init__(self, *args, **kwargs)
        self._on_css_updated()
        CSS.global_css.am_subscribe(self._on_css_updated)

        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()
        kernel_manager.kernel.gui = "qt4"

        if namespace is not None:
            self.push_namespace(namespace)

        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop() -> None:
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            guisupport.get_app_qt4().exit()

        self.exit_requested.connect(stop)

        self.enable_calltips = False

    def __del__(self) -> None:
        super().__del__()
        CSS.global_css.am_unsubscribe(self._on_css_updated)

    def _on_css_updated(self) -> None:
        self.setStyleSheet(CSS.global_css.am_obj)

    def push_namespace(self, namespace) -> None:
        self.kernel_manager.kernel.shell.push(namespace)

    def clear_terminal(self) -> None:
        self._control.clear()

    def print_text(self, text: str) -> None:
        self._append_plain_text(text, True)

    def execute_command(self, command) -> None:
        self._execute(command, False)



from __future__ import annotations

from PySide6.QtCore import QSize
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QToolBar

from .toolbar_action import ToolbarAction, ToolbarSplitter


class Toolbar:
    def __init__(self, window, name: str) -> None:
        self.window = window
        self.name = name

        self.actions = []
        self._cached: QToolBar | None = None
        self._cached_actions = {}

    def shutdown(self) -> None:
        """
        Prepare for deletion.
        """

    def qtoolbar(self):
        if self._cached is not None:
            return self._cached

        toolbar = QToolBar(self.name, self.window)

        for action in self.actions:
            if action in self._cached_actions:
                act = self._cached_actions[action]
            else:
                act = self._translate_element(toolbar, action)
                if act is not None:
                    self._cached_actions[action] = act

        toolbar.setIconSize(QSize(16, 16))

        self._cached = toolbar
        return toolbar

    @staticmethod
    def _translate_element(toolbar, action):
        if isinstance(action, ToolbarSplitter):
            toolbar.addSeparator()
            return None
        elif isinstance(action, ToolbarAction):
            if action.icon is not None:
                act = QAction(action.icon, action.name, toolbar)
            else:
                act = QAction(action.name, toolbar)
            if action.triggered is not None:
                act.triggered.connect(action.triggered)
            if action.tooltip:
                act.setToolTip(action.tooltip)
            if action.shortcut:
                act.setShortcuts(action.shortcut)
            act.setCheckable(action.checkable)
            toolbar.addAction(act)
            return act
        else:
            raise TypeError("Bad toolbar action", action)

    def add(self, element) -> None:
        if self._cached is not None:
            act = self._translate_element(self._cached, element)
            if act is not None:
                self._cached_actions[element] = act

        self.actions.append(element)

    def remove(self, element):
        # REQUIRES object identity
        try:
            act = self._cached_actions[element]
        except KeyError as ex:
            raise ValueError(f"Element {element} not found") from ex

        self.actions.remove(element)
        if self._cached is not None:
            self._cached.removeAction(act)
            del self._cached_actions[element]



from __future__ import annotations

import itertools
import re
from typing import TYPE_CHECKING

import networkx as nx
from angr.analyses.disassembly import ConstantOperand, Disassembly, Instruction, MemoryOperand
from angr.errors import SimEngineError

from angrmanagement.utils import string_at_addr

if TYPE_CHECKING:
    from angr.block import CapstoneInsn
    from angr.knowledge_plugins.functions.function import Function

    from angrmanagement.ui.views import DisassemblyView


class FunctionDiff:
    """
    A class to represent a diff of two functions, which may either be part of a larger disassembly view or not.
    Diffs are recorded in the rev_{change,add,del} sets, which are instruction addresses.

    """

    OBJ_DELETED = "del"
    OBJ_ADDED = "add"
    OBJ_CHANGED = "chg"
    OBJ_UNMODIFIED = "nop"

    # pylint:disable=unused-argument
    def __init__(
        self,
        func_base: Function,
        func_rev: Function,
        disas_base: Disassembly = None,
        disas_rev: Disassembly = None,
        prefer_symbols: bool = True,
        resolve_strings: bool = True,
        resolve_insn_addrs: bool = True,
        **kwargs,
    ) -> None:
        self.func_base = func_base
        self.func_rev = func_rev
        self.disas_base = disas_base
        self.disas_rev = disas_rev

        self.rev_change_set = set()
        self.rev_add_set = set()
        self.rev_del_set = set()

        self._prefer_symbols = prefer_symbols
        self._resolve_strings = resolve_strings
        self._resolve_insn_addrs = resolve_insn_addrs
        self._insn_mnem_check = 3

    @property
    def differs(self):
        return bool(self.rev_change_set or self.rev_add_set or self.rev_del_set)

    @property
    def prefer_symbols(self):
        return self._prefer_symbols and self.disas_base is not None and self.disas_rev is not None

    @staticmethod
    def is_executable_address(address, proj):
        for section in proj.loader.main_object.sections:
            if section.is_executable and section.min_addr <= address <= section.max_addr:
                return True

        return False

    def _linear_asm_from_function(
        self, func: Function, disas: Disassembly = None, as_dict: bool = False
    ) -> list[CapstoneInsn]:
        sorted_blocks = sorted(func.blocks, key=lambda b: b.addr)
        instruction_lists = [block.disassembly.insns for block in sorted_blocks]
        instructions = list(itertools.chain.from_iterable(instruction_lists))
        if not self.prefer_symbols:
            return instructions if not as_dict else {i.address: i for i in instructions}

        symbolized_instructions = [disas.raw_result_map["instructions"][insn.address] for insn in instructions]

        return symbolized_instructions if not as_dict else {i.addr: i for i in symbolized_instructions}

    def diff_insn(self, base_insn: CapstoneInsn | Instruction, rev_insn: CapstoneInsn | Instruction):
        if self._prefer_symbols:
            if base_insn.render() == rev_insn.render():
                return FunctionDiff.OBJ_UNMODIFIED

            if base_insn.mnemonic.render() == rev_insn.mnemonic.render() and len(base_insn.operands) == len(
                rev_insn.operands
            ):
                if not self._resolve_strings and not self._resolve_insn_addrs:
                    return FunctionDiff.OBJ_CHANGED

                # attempt to resolve strings that act as symbols
                base_mem_ops = [op for op in base_insn.operands if isinstance(op, MemoryOperand | ConstantOperand)]
                if base_mem_ops:
                    rev_mem_ops = [op for op in rev_insn.operands if isinstance(op, MemoryOperand | ConstantOperand)]
                    if len(rev_mem_ops) == len(base_mem_ops) == 1:
                        base_mem_op = base_mem_ops[0]
                        rev_mem_op = rev_mem_ops[0]

                        try:
                            if isinstance(base_mem_op, MemoryOperand):
                                base_mem_op_addr = list(base_mem_op.values)[0].val
                            else:
                                base_mem_op_addr = base_mem_op.children[0].val
                        except (IndexError, KeyError, ValueError, AttributeError):
                            return FunctionDiff.OBJ_CHANGED

                        try:
                            if isinstance(rev_mem_op, MemoryOperand):
                                rev_mem_op_addr = list(rev_mem_op.values)[0].val
                            else:
                                rev_mem_op_addr = rev_mem_op.children[0].val
                        except (IndexError, KeyError, ValueError, AttributeError):
                            return FunctionDiff.OBJ_CHANGED

                        base_str = string_at_addr(
                            self.func_base.project.kb.cfgs.get_most_accurate(), base_mem_op_addr, self.func_base.project
                        )
                        rev_str = string_at_addr(
                            self.func_rev.project.kb.cfgs.get_most_accurate(), rev_mem_op_addr, self.func_rev.project
                        )

                        if base_str is not None and base_str == rev_str:
                            base_insn_str = re.sub(r"\[.*\]", base_str, base_insn.render()[0])
                            rev_insn_str = re.sub(r"\[.*\]", base_str, rev_insn.render()[0])

                            if base_insn_str == rev_insn_str:
                                return FunctionDiff.OBJ_UNMODIFIED
                        elif self._resolve_insn_addrs and (
                            self.is_executable_address(base_mem_op_addr, self.func_base.project)
                            and self.is_executable_address(rev_mem_op_addr, self.func_rev.project)
                        ):
                            try:
                                base_blk = self.func_base.get_block(base_mem_op_addr)
                            except SimEngineError:
                                return FunctionDiff.OBJ_CHANGED

                            try:
                                rev_blk = self.func_rev.get_block(rev_mem_op_addr)
                            except SimEngineError:
                                return FunctionDiff.OBJ_CHANGED

                            base_mnem = [ins.mnemonic for ins in base_blk.disassembly.insns][: self._insn_mnem_check]
                            rev_mnem = [ins.mnemonic for ins in rev_blk.disassembly.insns][: self._insn_mnem_check]
                            if base_mnem == rev_mnem:
                                return FunctionDiff.OBJ_UNMODIFIED

                return FunctionDiff.OBJ_CHANGED

            return FunctionDiff.OBJ_ADDED
        else:
            if base_insn.mnemonic != rev_insn.mnemonic or len(base_insn.operands) != len(rev_insn.operands):
                return FunctionDiff.OBJ_ADDED

            if base_insn.op_str != rev_insn.op_str:
                return FunctionDiff.OBJ_CHANGED

            return FunctionDiff.OBJ_UNMODIFIED

    def addr_diff_value(self, addr: int):
        if addr in self.rev_del_set:
            return self.OBJ_DELETED
        elif addr in self.rev_add_set:
            return self.OBJ_ADDED
        elif addr in self.rev_change_set:
            return self.OBJ_CHANGED
        else:
            return self.OBJ_UNMODIFIED

    def compute_function_diff(self) -> None:
        pass


class LinearFunctionDiff(FunctionDiff):
    """
    A function diff calculated by traversing two disassemblies linearly and checking index by index.
    """

    def __init__(
        self,
        func_base: Function,
        func_rev: Function,
        disas_base: Disassembly = None,
        disas_rev: Disassembly = None,
        prefer_symbols: bool = True,
        resolve_strings: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            func_base,
            func_rev,
            prefer_symbols=prefer_symbols,
            resolve_strings=resolve_strings,
            disas_base=disas_base,
            disas_rev=disas_rev,
            **kwargs,
        )
        self.base_insns = self._linear_asm_from_function(func_base, disas=self.disas_base)
        self.rev_insns = self._linear_asm_from_function(func_rev, disas=self.disas_rev)
        self.compute_function_diff()

    def compute_function_diff(self) -> None:
        for idx, base_insn in enumerate(self.base_insns):
            if idx >= len(self.rev_insns):
                break

            rev_insn = self.rev_insns[idx]
            rev_addr = rev_insn.addr if self.prefer_symbols else rev_insn.address
            if self.diff_insn(base_insn, rev_insn) in (self.OBJ_CHANGED, self.OBJ_ADDED):
                self.rev_change_set.add(rev_addr)
                continue

        if len(self.rev_insns) <= len(self.base_insns) or not self.base_insns:
            return

        # if we end up here, then we have more rev_insns to parse
        # pylint:disable=undefined-loop-variable
        for rev_insn in self.rev_insns[idx:]:
            rev_addr = rev_insn.addr if self.prefer_symbols else rev_insn.address
            self.rev_add_set.add(rev_addr)


class BFSFunctionDiff(FunctionDiff):
    """
    Use two graphs to compute a function diff, performing Linear diff on each block for each position.
    The traversal of both graphs are done in a BFS manner.
    """

    def __init__(
        self,
        func_base: Function,
        func_rev: Function,
        view_base: DisassemblyView = None,
        view_rev: DisassemblyView = None,
        **kwargs,
    ) -> None:
        super().__init__(func_base, func_rev, **kwargs)
        self.base_cfg = view_base._flow_graph.function_graph.supergraph
        self.rev_cfg = view_rev._flow_graph.function_graph.supergraph
        self.base_insns = self._linear_asm_from_function(func_base, disas=self.disas_base, as_dict=True)
        self.rev_insns = self._linear_asm_from_function(func_rev, disas=self.disas_rev, as_dict=True)
        self.compute_function_diff()

    @staticmethod
    def supergraph_block_to_insns(function, super_block):
        instructions = []
        for cnode in super_block.cfg_nodes:
            ins_blk = function.get_block(cnode.addr)
            instructions += list(ins_blk.disassembly.insns)

        return instructions

    @staticmethod
    def bfs_list_block_levels(graph: nx.DiGraph):
        block_levels = []
        start_block = [n for n in graph.nodes if graph.in_degree(n) == 0][0]
        bfs = list(nx.bfs_successors(graph, start_block))
        for blk_tree in bfs:
            source, children = blk_tree

            if len(children) == 1:
                block_levels.append(children)
            elif len(children) == 2:
                fallthrough = source.addr + source.size
                if children[0].addr == fallthrough:
                    first, second = children[:]
                else:
                    first, second = children[::-1]

                block_levels.append([first, second])

        block_levels = [[start_block]] + block_levels
        return block_levels

    def compute_function_diff(self) -> None:
        base_levels = self.bfs_list_block_levels(self.base_cfg)
        rev_levels = self.bfs_list_block_levels(self.rev_cfg)
        diff_map = {}

        for level_idx, base_level in enumerate(base_levels):
            if level_idx >= len(rev_levels):
                break

            rev_blocks = rev_levels[level_idx]
            for block_idx, base_block in enumerate(base_level):
                if block_idx >= len(rev_blocks):
                    break

                unmodified_insns = []
                rev_insns = self.supergraph_block_to_insns(self.func_rev, rev_blocks[block_idx])

                # find unmodified instructions
                for base_insn in self.supergraph_block_to_insns(self.func_base, base_block):
                    # we are now in a single block starting with the base instruction;
                    # what we want to do now is first find all the original instructions in the
                    # new block and mark those as nops
                    for rev_insn in rev_insns:
                        if rev_insn.address in diff_map:
                            continue

                        if (
                            self.diff_insn(self.base_insns[base_insn.address], self.rev_insns[rev_insn.address])
                            == self.OBJ_UNMODIFIED
                        ):
                            diff_map[rev_insn.address] = self.OBJ_UNMODIFIED
                            unmodified_insns.append(rev_insn.address)
                            break

                # find changed instructions
                first_unmodified_address = unmodified_insns[0] if unmodified_insns else None
                for insn_idx, _base_insn in enumerate(self.supergraph_block_to_insns(self.func_base, base_block)):
                    if insn_idx >= len(rev_insns):
                        break

                    rev_insn = rev_insns[insn_idx]
                    if first_unmodified_address is not None and rev_insn.address < first_unmodified_address:
                        diff_map[rev_insn.address] = self.OBJ_ADDED
                        continue

                    if rev_insn.address in diff_map:
                        continue

                    diff_map[rev_insn.address] = self.OBJ_CHANGED

        for rev_level in rev_levels:
            for rev_block in rev_level:
                for rev_insn in self.supergraph_block_to_insns(self.func_rev, rev_block):
                    if rev_insn.address not in diff_map:
                        diff_map[rev_insn.address] = self.OBJ_ADDED

        # add to colors
        for addr, diff_val in diff_map.items():
            if diff_val == self.OBJ_CHANGED:
                self.rev_change_set.add(addr)
            elif diff_val == self.OBJ_ADDED:
                self.rev_add_set.add(addr)



from __future__ import annotations

from collections import defaultdict


class VulnerabilityType:
    PATH_TRAVERSAL = 1
    COMMAND_INJECTION = 2
    SQL_INJECTION = 3

    @staticmethod
    def to_string(vuln_type: int) -> str:
        if vuln_type == VulnerabilityType.PATH_TRAVERSAL:
            return "Path travesal"
        elif vuln_type == VulnerabilityType.COMMAND_INJECTION:
            return "Command injection"
        elif vuln_type == VulnerabilityType.SQL_INJECTION:
            return "SQL injection"

        return "Unknown"


class BaseSink:
    pass


class FunctionArgumentSink(BaseSink):
    def __init__(self, lib: str | None, func_name: str, arg_idx, platforms: list[str] | None = None) -> None:
        self.lib = lib
        self.func_name = func_name
        self.arg_idx = arg_idx
        self.platforms = platforms

    def __repr__(self) -> str:
        return "<TS %s:arg %d>" % (self.func_name, self.arg_idx)


class SinkManager:
    def __init__(self, sinks: dict[int, list[BaseSink]]) -> None:
        self.sinks = sinks

        # lookup caches
        self._func_to_sinks: dict[str, list[tuple[int, FunctionArgumentSink]]] = defaultdict(list)

        self._init_caches()

    def _init_caches(self) -> None:
        for vuln_type, sinks in self.sinks.items():
            for sink in sinks:
                if isinstance(sink, FunctionArgumentSink):
                    self._func_to_sinks[sink.func_name].append((vuln_type, sink))

    def has_function_sink(self, func_name: str, lib: str | None = None) -> bool:
        if func_name not in self._func_to_sinks:
            return False
        if lib is None:
            return True
        # compare lib name
        sinks = self._func_to_sinks[func_name]
        return any(sink[1].lib == lib for sink in sinks)

    def get_function_sinks(self, func_name: str, lib: str | None = None) -> list[tuple[int, FunctionArgumentSink]]:
        if func_name not in self._func_to_sinks:
            return []
        if lib is None:
            return self._func_to_sinks[func_name]
        # compare lib name
        sinks = self._func_to_sinks[func_name]
        return [sink for sink in sinks if sink[1].lib == lib]


FAS = FunctionArgumentSink

_path_traversal_sinks = [
    FAS("libc", "fopen", 0),
    FAS("libc", "chdir", 0),
    FAS(None, "std::basic_ifstream::__ctor__", 1),
]

_command_injection_sinks = [
    FAS("libc", "system", 0),
    FAS("libc", "popen", 0),
]

# TODO: SQL injection sinks require certain conditions to be met
_sql_injection_sinks = [
    FAS("libc", "asprintf", 0),
    FAS("libc", "sprintf", 0),
]

_sinks = {
    VulnerabilityType.PATH_TRAVERSAL: _path_traversal_sinks,
    VulnerabilityType.COMMAND_INJECTION: _command_injection_sinks,
}

sink_manager = SinkManager(_sinks)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

import angr
from PySide6.QtCore import Qt
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QAbstractItemView, QMenu, QTableWidget, QTableWidgetItem

from angrmanagement.ui.dialogs.new_state import NewState
from angrmanagement.utils.namegen import NameGenerator

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class QStateTableItem(QTableWidgetItem):
    """
    An entry within a QStateTable
    """

    def __init__(self, state) -> None:
        super().__init__()

        self.state = state

    def widgets(self):
        state = self.state

        name = state.gui_data.name
        base_name = state.gui_data.base_name
        is_changed = "No" if state.gui_data.is_original else "Yes"
        mode = state.mode
        address = f"{state.addr:x}" if isinstance(state.addr, int) else "Symbolic"
        state_options = {o for o, v in state.options._options.items() if v is True}
        options_plus = state_options - angr.sim_options.modes[mode]
        options_minus = angr.sim_options.modes[mode] - state_options
        options = " ".join([" ".join("+" + o for o in options_plus), " ".join("-" + o for o in options_minus)])

        widgets = [
            QTableWidgetItem(name),
            QTableWidgetItem(address),
            QTableWidgetItem(is_changed),
            QTableWidgetItem(base_name),
            QTableWidgetItem(mode),
            QTableWidgetItem(options),
        ]

        if state.gui_data.is_base:
            color = QColor(0, 0, 0x80)
        elif state.gui_data.is_original:
            color = QColor(0, 0x80, 0)
        else:
            color = QColor(0, 0, 0)

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)
            w.setForeground(color)

        return widgets


class QStateTable(QTableWidget):
    """
    The table which is the subject of the States View
    """

    def __init__(self, workspace: Workspace, instance: Instance, parent, selection_callback=None) -> None:
        super().__init__(parent)

        self._selected = selection_callback

        header_labels = ["Name", "Address", "Changed?", "Base State", "Mode", "Options"]

        self.setColumnCount(len(header_labels))
        self.setHorizontalHeaderLabels(header_labels)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.items = []
        self.workspace = workspace
        self.instance = instance
        self.states = instance.states

        self.itemDoubleClicked.connect(self._on_state_selected)
        self.cellDoubleClicked.connect(self._on_state_selected)
        self.states.am_subscribe(self._watch_states)
        self.reload()

    def closeEvent(self, _) -> None:
        self.states.am_unsubscribe(self._watch_states)

    def current_state_record(self):
        selected_index = self.currentRow()
        if 0 <= selected_index < len(self.items):
            return self.items[selected_index]
        else:
            return None

    def reload(self) -> None:
        # current_row = self.currentRow()
        self.clearContents()

        self.items = [QStateTableItem(f) for f in self.states]
        items_count = len(self.items)
        self.setRowCount(items_count)

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        # if 0 <= current_row < len(self.items):
        #    self.setCurrentItem(current_row, 0)

    def _on_state_selected(self, *args) -> None:  # pylint: disable=unused-argument
        if self._selected is not None:
            self._selected(self.current_state_record())

    def contextMenuEvent(self, event) -> None:
        sr = self.current_state_record()

        menu = QMenu("", self)

        menu.addAction("New state...", self._action_new_state)
        menu.addSeparator()

        a = menu.addAction("Duplicate state", self._action_duplicate)
        if sr is None:
            a.setDisabled(True)

        a = menu.addAction("Delete state", self._action_delete)
        if sr is None:
            a.setDisabled(True)

        a = menu.addAction("New simulation manager", self._action_new_simulation_manager)
        if sr is None:
            a.setDisabled(True)

        menu.exec_(event.globalPos())

    def _action_new_state(self) -> None:
        dialog = NewState(self.workspace, self.instance, parent=self)
        dialog.exec_()

    def _action_duplicate(self) -> None:
        state = self.states[self.currentRow()]
        copy = state.copy()
        copy.gui_data.name = self._get_copied_state_name(copy.gui_data.name)
        self.states.append(copy)
        self.states.am_event(src="duplicate", state=copy)

    def _action_delete(self) -> None:
        tmp = self.states.pop(self.currentRow())
        self.states.am_event(src="delete", state=tmp)

    def _action_new_simulation_manager(self) -> None:
        state = self.states[self.currentRow()]
        simgr_name = NameGenerator.random_name()
        self.workspace.create_simulation_manager(state, simgr_name)

    def _watch_states(self) -> None:
        self.reload()

    def _get_copied_state_name(self, current_name: str):
        """
        Get a non-duplicating name for the copied state.

        :param str current_name:    The current name of the state.
        :return:                    A new name of the copied state.
        :rtype:                     str
        """

        m = re.match(r"^([\s\S]*) copy\s*(\d*)$", current_name)

        if m:
            # ends with copy
            ctr_str = m.group(2)
            ctr = int(ctr_str) + 1 if ctr_str else 1

            current_name = m.group(1)
            name = current_name + " copy %d" % ctr
        else:
            ctr = 0
            name = current_name + " copy"

        # Increment the counter until there is no conflict with existing names
        all_names = {s.gui_data.name for s in self.states}
        while name in all_names:
            ctr += 1
            name = current_name + " copy %d" % ctr
        return name



# pylint:disable=import-outside-toplevel,unused-import,no-member
from __future__ import annotations

import asyncio
import ctypes
import multiprocessing
import os
import signal
import sys
import threading
import time
import warnings

from . import __version__

if sys.platform.startswith("darwin"):
    from Foundation import NSBundle  # pylint: disable=import-error


def shut_up(*args, **kwargs) -> None:  # pylint:disable=unused-argument
    return


warnings.simplefilter = shut_up


name: str = "angr management"


def set_app_user_model_id() -> None:
    # Explicitly call SetCurrentProcessExplicitAppUserModelID() so the taskbar icon is displayed correctly.

    if sys.platform == "win32":
        winver = sys.getwindowsversion()
        if winver.major >= 5:
            myappid = name.replace(" ", "-")
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)


def set_windows_event_loop_policy() -> None:
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def start_management(filepath=None, use_daemon=None, profiling: bool = False) -> None:
    set_app_user_model_id()
    set_windows_event_loop_policy()

    from PySide6.QtCore import QMargins, QRectF, Qt
    from PySide6.QtGui import QCursor, QFontDatabase, QGuiApplication, QIcon, QPixmap
    from PySide6.QtWidgets import QApplication, QSplashScreen

    from .config import FONT_LOCATION, IMG_LOCATION, Conf

    class SplashScreen(QSplashScreen):
        """
        angr-management splash screen, showing version, a progress bar, and progress status message.

        Note: Progress message is distinct from the one provided by QSplashScreen::showMessage.
        """

        _progress: float = 0.0
        _progress_message: str = ""

        def setProgress(self, progress: float, progress_message: str | None = None) -> None:
            self._progress = progress
            if self._progress_message is not None:
                self._progress_message = progress_message
            self.repaint()

        def drawContents(self, painter) -> None:
            super().drawContents(painter)
            contentsRect = self.contentsRect()

            # Draw progress bar
            pbar_height = 3
            pbar_width = contentsRect.width() * max(0.0, min(self._progress, 1.0))
            painter.setPen(Qt.transparent)
            painter.setBrush(Qt.white)
            painter.drawRect(QRectF(0, contentsRect.height() - pbar_height, pbar_width, pbar_height))

            # Draw version and status text
            pad = 6
            r = contentsRect.marginsRemoved(QMargins(pad, pad, pad, pad + pbar_height))
            painter.setPen(Qt.white)
            painter.drawText(r, Qt.AlignTop | Qt.AlignRight, __version__)
            painter.drawText(r, Qt.AlignBottom | Qt.AlignLeft, self._progress_message)

    # Fix app title on macOS
    if sys.platform.startswith("darwin"):
        try:
            bundle = NSBundle.mainBundle()
            info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
            info["CFBundleName"] = name
        except Exception as e:  # pylint: disable=broad-except
            # This happens before logging is setup so use stderr
            print(f"Failed to set App name! {type(e).__name__}: {e}", file=sys.stderr)

    app = QApplication(sys.argv)
    app.setApplicationDisplayName(name)
    app.setApplicationName(name)
    icon_location = os.path.join(IMG_LOCATION, "angr.png")
    QApplication.setWindowIcon(QIcon(icon_location))

    # try to import the initial configuration for the install
    Conf.attempt_importing_initial_config()

    # Make + display splash screen
    splashscreen_location = os.path.join(IMG_LOCATION, "angr-splash.png")
    splash_pixmap = QPixmap(splashscreen_location)
    current_screen = QGuiApplication.screenAt(QCursor.pos())
    splash = SplashScreen(current_screen, splash_pixmap, Qt.WindowStaysOnTopHint)
    splash.setFixedSize(splash_pixmap.size())

    icon_location = os.path.join(IMG_LOCATION, "angr.png")
    splash.setWindowIcon(QIcon(icon_location))
    splash.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.FramelessWindowHint)
    splash.setEnabled(False)
    splash.show()
    for _ in range(5):
        time.sleep(0.01)
        app.processEvents()

    splash.setProgress(0.1, "Importing modules")
    import angr

    from .logic import GlobalInfo
    from .ui.awesome_tooltip_event_filter import QAwesomeTooltipEventFilter
    from .ui.css import refresh_theme  # import .ui after showing the splash screen since it's going to take time
    from .ui.main_window import MainWindow

    angr.loggers.profiling_enabled = bool(profiling)

    splash.setProgress(0.5, "Configuring theme")
    refresh_theme()

    # Load fonts, initialize font-related configuration
    QFontDatabase.addApplicationFont(os.path.join(FONT_LOCATION, "SourceCodePro-Regular.ttf"))
    QFontDatabase.addApplicationFont(os.path.join(FONT_LOCATION, "DejaVuSansMono.ttf"))
    Conf.init_font_config()
    Conf.connect("ui_default_font", app.setFont, True)

    # install the global tooltip filter
    app.installEventFilter(QAwesomeTooltipEventFilter(app))

    splash.setProgress(0.9, "Initializing main window")
    GlobalInfo.gui_thread = threading.get_ident()
    file_to_open = filepath if filepath else None
    main_window = MainWindow(app=app, use_daemon=use_daemon)
    splash.setProgress(1.0, "")
    splash.finish(main_window)

    if file_to_open is not None:
        main_window.load_file(file_to_open)

    main_window.initialized = True
    main_window.workspace.view_manager.main_window_initialized()

    if file_to_open is None:
        main_window.show_welcome_dialog()

    app.exec_()


def main() -> None:
    import argparse

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    prog: str = name.replace(" ", "-")
    parser = argparse.ArgumentParser(prog=prog, description=name)
    parser.add_argument("-v", "--version", action="version", version=f"{prog} {__version__}")
    parser.add_argument("-s", "--script", type=str, help="run a python script in the (commandline) angr environment")
    parser.add_argument("-i", "--interactive", action="store_true", help="interactive (ipython) mode")
    parser.add_argument("-n", "--no-gui", action="store_true", help="run in headless mode")
    parser.add_argument(
        "-d",
        "--with-daemon",
        action="store_true",
        help="use angr with the daemon. this allows angr "
        "to handle angr:// URLs. it will "
        "automatically start a daemon if there isn't "
        "already one running.",
    )
    parser.add_argument("-D", "--daemon", action="store_true", help="start a daemon to handle angr:// URLs.")
    parser.add_argument(
        "-u", "--url", type=str, nargs="?", help="(internal) handle angr:// URLs. the daemon must be running."
    )
    parser.add_argument("-p", "--profiling", action="store_true", help="display profiling log messages.")
    parser.add_argument("-R", "--autoreload", action="store_true", help="Reload all python modules on each job start.")
    parser.add_argument("binary", nargs="?", help="the binary to open (for the GUI)")

    args = parser.parse_args()

    if args.autoreload:
        from .logic import GlobalInfo

        GlobalInfo.autoreload = True

    if args.daemon:
        from .daemon import start_daemon

        start_daemon()
        return
    elif args.url:
        from .daemon import daemon_conn, daemon_exists, handle_url, run_daemon_process

        if not daemon_exists():
            run_daemon_process()
            time.sleep(1)

        # initialize plugins
        from .plugins import PluginManager

        PluginManager(None).discover_and_initialize_plugins()

        action = handle_url(args.url, act=False)
        action.act(daemon_conn())
        return
    if args.script:
        import runpy

        runpy.run_path(args.script)  # pylint:disable=unused-variable
    if args.interactive:
        if args.script:
            print("Your script's globals() dict is available in the `script_globals` variable.")
        import IPython

        IPython.embed(banner1="")
    if not args.no_gui:
        start_management(
            args.binary, use_daemon=True if args.with_daemon else None, profiling=True if args.profiling else None
        )


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



from __future__ import annotations

import os

from PySide6.QtGui import QIcon

from angrmanagement.config import IMG_LOCATION

from .toolbar import Toolbar, ToolbarAction


class FunctionTableToolbar(Toolbar):
    def __init__(self, function_table) -> None:
        super().__init__(function_table, "Function table options")

        # TODO: An icon would be great
        self._alignment_action = ToolbarAction(
            QIcon(os.path.join(IMG_LOCATION, "toolbar-show-alignment.png")),
            "Show alignment functions",
            "Display alignment function stubs.",
            function_table.toggle_show_alignment_functions,
            checkable=True,
        )

        self.actions = [
            self._alignment_action,
        ]

    def toggle_show_alignment_functions(self) -> None:
        self.window.toggle_show_alignment_functions()

        if self._cached_actions and self._alignment_action in self._cached_actions:
            self._cached_actions[self._alignment_action].setChecked(self.window.show_alignment_functions)
